/****************************************************************************
*                                                                           *
*                  ___  ____  ____   ___  __    ____  ____                  *
*                 / __)(_  _)(  _ \ / __)(  )  ( ___)(  _ \                 *
*                ( (__  _)(_  )   /( (__  )(__  )__)  )(_) )                *
*                 \___)(____)(_)\_) \___)(____)(____)(____/                 *
*                                                                           *
*                                                                           *
*                                                                           *
*   This file is part of Circled                                            *
*                                                                           *
*   Copyright (C) 2011  Fabien LOISON                                       *
*                                                                           *
*   Circled is free software: you can redistribute it and/or modify         *
*   it under the terms of the GNU General Public License as published by    *
*   the Free Software Foundation, either version 3 of the License, or       *
*   (at your option) any later version.                                     *
*                                                                           *
*   This program is distributed in the hope that it will be useful,         *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*   GNU General Public License for more details.                            *
*                                                                           *
*   You should have received a copy of the GNU General Public License       *
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
*                                                                           *
****************************************************************************/


/**
 * This file contains all the class/functions for the Circled game.
 */


#include "_headers.h"


//
// Board
//


/**
 * The constructor of the Board class.
 *
 * Initialize an new board.
 *
 * Args:
 *     terminal -- The Terminal instance.
 *     w -- The width of the board.
 *     h -- The height of the board.
 */
Board::Board(Terminal *terminal, int w, int h)
{
	term = terminal;

	//Check width
	if ( w < 6 || w % 2 != 0)
	{
		width = 6;
	}
	else
	{
		width = w;
	}

	//Check height
	if ( h < 5 || h % 2 == 0)
	{
		height = 5;
	}
	else
	{
		height = h;
	}

	//Initialize the board
	int x;
	int y;
	board = new int* [height];
	for (y=0 ; y<height ; y++)
	{
		board[y] = new int [width];
		for (x=0 ; x<width ; x++)
		{
			set_cell(x, y, 0);
		}
	}

	//Calculate the board position
	pos_x = (term->get_width() - width*2) / 2 - 1;
	pos_y = (term->get_height() - height) /2;

	//Initialize the MouseEventBox
	evbox = MouseEventBox(pos_x, pos_y, pos_x + width * 2, pos_y + height);
}

/**
 * Draw the board on the terminal.
 */
void Board::refresh()
{
	//Draw the board shadow
	term->set_bg_color(COLOR_BLACK);
	term->draw_rectangle(pos_x + 2, pos_y + 1,
			pos_x + width * 2 + 2, pos_y + height + 1);

	//Draw the board
	int x;
	int y;
	for (y=0 ; y<height ; y++)
	{
		term->move_xy(pos_x, pos_y + y);
		for (x=0 ; x<width ; x++)
		{
			switch (get_cell(x, y))
			{
				case CELL_EMPTY:
					if ((x + y) % 2 == 1)
					{
						term->set_bg_color(COLOR_BOARD_1);
					}
					else
					{
						term->set_bg_color(COLOR_BOARD_2);
					}
					cout << "  " << flush;
					break;
					
				case CELL_P1:
					term->set_fg_color(COLOR_WHITE);
					term->set_bg_color(COLOR_PLAYER_1);
					cout << "P1" << flush;
					break;

				case CELL_P2:
					term->set_fg_color(COLOR_WHITE);
					term->set_bg_color(COLOR_PLAYER_2);
					cout << "P2" << flush;
					break;

				case CELL_BRICK:
					term->set_bg_color(COLOR_BRICK);
					cout << "  " << flush;
					break;
					
			}
		}
	}
}

/**
 * Get the value of the given cell.
 *
 * Args:
 *     x, y -- The coordinate of the cell.
 *
 * Returns:
 *     The value of the cell.
 */
int Board::get_cell(int x, int y)
{
	if (x >= 0 && x < width && y >= 0 && y < height)
	{
		return board[y][x];
	}
	else
	{
		return CELL_BRICK;
	}
}

/**
 * Set the value of the given cell.
 *
 * Args:
 *     x, y -- The coordinate of the cell.
 *     value -- The new value of the cell (CELL_EMPTY, CELL_P1,
 *              CELL_P2 or CELL_BRICK).
 */
void Board::set_cell(int x, int y, int value)
{
	board[y][x] = value;
}

/**
 * Check if the cell is free.
 *
 * Returns:
 *     true if the cell is free, false else.
 */
bool Board::cell_is_free(int x, int y)
{
	if (get_cell(x, y) == CELL_EMPTY)
	{
		return true;
	}
	else
	{
		return false;
	}
}

/**
 * Get the board width.
 *
 * Returns:
 *     The board width.
 */
int Board::get_width()
{
	return width;
}

/**
 * Get the board height.
 *
 * Returns:
 *     The board height.
 */
int Board::get_height()
{
	return height;
}


//
// Piece
//


/**
 * The contructor of the Piece class.
 *
 * Args:
 *     x, y -- The coordinates of the piece.
 */
Piece::Piece(int x, int y)
{
	set_x(x);
	set_y(y);
}

/**
 * Get the x position of the piece.
 *
 * Returns:
 *     The x position of the piece.
 */
int Piece::get_x()
{
	return pos_x;
}

/**
 * Set the x position of the piece.
 *
 * Args:
 *     x -- The new x position.
 */
void Piece::set_x(int x)
{
	pos_x = x;
}

/**
 * Get the y position of the piece.
 *
 * Returns:
 *     The y position of the piece.
 */
int Piece::get_y()
{
	return pos_y;
}

/**
 * Set the y position of the piece.
 *
 * Args:
 *     y -- The new y position.
 */
void Piece::set_y(int y)
{
	pos_y = y;
}


//
// Player
//


/**
 * The constructor of the Player class.
 *
 * Args:
 *     x, y -- The initial position of the player on the board.
 *     playerid -- The player id (PLAYER_1 or PLAYER_2).
 *     nick -- The nickname of the player.
 */
Player::Player(int x, int y, int playerid, std::string nick)
{
	piece = Piece(x, y);
	nickname = nick;
	id = playerid;
}

/**
 * Initialize the player on a new board.
 *
 * Args:
 *     board -- The instance of the board.
 */
void Player::init(Board *board)
{
	board->set_cell(piece.get_x(), piece.get_y(), id);
	board->refresh();
}

/**
 * Check if the piece of the player can move.
 *
 * Args:
 *     board -- The instance of the board.
 *
 * Returns:
 *     true if the piece can move, false else.
 */
bool Player::can_move(Board *board)
{
	int x;
	int y;
	for (y=piece.get_y()-1 ; y<=piece.get_y()+1 ; y++)
	{
		for (x=piece.get_x()-1 ; x<=piece.get_x()+1 ; x++)
		{
			if (board->cell_is_free(x, y))
			{
				return true;
			}
		}
	}
	return false;
}

/**
 * Ask the player for moving its piece.
 *
 * Args:
 *     board -- The instance of the board.
 *
 * Returns:
 *     true if the user abort the game, false else.
 */
bool Player::move_piece(Board *board)
{
	//Display the message to the player
	display_hint("Move your piece to another cell.", board->term);

	//Wait for input
	bool valid_input = false;
	Event ev;
	do {
		ev = board->term->get_event();
		if (ev.get_type() == EV_TYPE_MOUSE && board->evbox.check_collide(&ev))
		{
			int x = board->evbox.get_relative_x(&ev) / 2;
			int y = board->evbox.get_relative_y(&ev);
			if (x >= piece.get_x() - 1 && x <= piece.get_x() + 1 &&
			    y >= piece.get_y() - 1 && y <= piece.get_y() + 1 &&
			    board->cell_is_free(x, y))
			{
				board->set_cell(piece.get_x(), piece.get_y(), CELL_EMPTY);
				board->set_cell(x, y, id);
				piece.set_x(x);
				piece.set_y(y);
				valid_input = true;
			}
		}
		else if (ev.get_type() == EV_TYPE_CHAR && ev.get_char() == 'q')
		{
			return true;
		}
	} while (!valid_input);
	board->refresh();
	return false;
}

/**
 * Ask the player for a new brick.
 *
 * Args:
 *     board -- The instance of the board.
 *
 * Returns:
 *     true if the user abort the game, false else.
 */
bool Player::put_brick(Board *board)
{
	//Display the message to the player
	display_hint("Put a brick on any empty cell.", board->term);

	//Wait for input
	bool valid_input = false;
	Event ev;
	do {
		ev = board->term->get_event();
		if (ev.get_type() == EV_TYPE_MOUSE && board->evbox.check_collide(&ev))
		{
			int x = board->evbox.get_relative_x(&ev) / 2;
			int y = board->evbox.get_relative_y(&ev);
			if (board->cell_is_free(x, y))
			{
				board->set_cell(x, y, CELL_BRICK);
				valid_input = true;
			}
		}
		else if (ev.get_type() == EV_TYPE_CHAR && ev.get_char() == 'q')
		{
			return true;
		}
	} while (!valid_input);
	board->refresh();
	return false;
}

/**
 * Get the player nickname.
 *
 * Returns:
 *     The player nickname.
 */
std::string Player::get_nickname()
{
	return nickname;
}

/**
 * Display a message to the player.
 *
 * Args:
 *     hint -- The message to display.
 *     term -- The Terminal instance.
 */
void Player::display_hint(char *hint, Terminal *term)
{
	//Clear the hint bar
	term->set_bg_color(COLOR_UI);
	term->draw_rectangle(0, term->get_height() - 1,
			term->get_width(),  term->get_height());
	term->move_xy(1, term->get_height() - 1);
	term->set_fg_color(COLOR_WHITE);
	if (id == PLAYER_1)
	{
		term->set_bg_color(COLOR_PLAYER_1);
		cout << "P1" << flush;
	}
	else
	{
		term->set_bg_color(COLOR_PLAYER_2);
		cout << "P2" << flush;
	}
	term->set_bg_color(COLOR_UI);
	cout << " " << hint << flush;
}


//
// Game
//


/**
 * The constructor of the Game class.
 *
 * Initializes the game.
 *
 * Args:
 *     width -- The width of the board.
 *     height -- The height of the board.
 *     nick_p1 -- The nickname of the first player.
 *     nick_p2 -- The nickname of the second player.
 *     term -- The instance of the Terminal.
 */
Game::Game(int width, int height, std::string nick_p1,
           std::string nick_p2, Terminal *term)
{
	//Board
	board = Board(term, width, height);

	//Player 1
	players[0] = Player(
			board.get_width() / 2 - 1, //x
			board.get_height() / 2,    //y
			PLAYER_1, nick_p1);
	players[0].init(&board);

	//Player 2
	players[1] = Player(
			board.get_width() / 2,  //x
			board.get_height() / 2, //y
			PLAYER_2, nick_p2);
	players[1].init(&board);

	winner = -1;
	in_game = true;
}

/**
 * Check if the game is finished.
 *
 * If a player win, in_game will be set to false, and the index of the winner
 * will be reported in winner.
 */
void Game::check_end_of_game()
{
	if (!players[1].can_move(&board))
	{
		in_game = false;
		winner = 0;
		return;
	}

	if (!players[0].can_move(&board))
	{
		in_game = false;
		winner = 1;
		return;
	}
}


//
// functions
//


/**
 * Display the logo/copyright screen with fade-in and fade-out effect.
 *
 * Args:
 *     term -- The Terminal instance.
 */
void start_screen(Terminal *term)
{
	term->set_bg_color(COLOR_BLACK);
	term->draw_rectangle(0, 0, term->get_width(), term->get_height());
	int banner_x = (term->get_width() - 43) / 2;
	int banner_y = (term->get_height() - 4) / 2.5;
	int copyright_x = (term->get_width() - 33) / 2;
	int color;
	for (color=232 ; color<=255 ; color++)
	{
		term->set_fg_color(color);
		term->print_xy(banner_x, banner_y + 0, "  ___  ____  ____   ___  __    ____  ____  ");
		term->print_xy(banner_x, banner_y + 1, " / __)(_  _)(  _ \\ / __)(  )  ( ___)(  _ \\ ");
		term->print_xy(banner_x, banner_y + 2, "( (__  _)(_  )   /( (__  )(__  )__)  )(_) )");
		term->print_xy(banner_x, banner_y + 3, " \\___)(____)(_)\\_) \\___)(____)(____)(____/ ");
		term->print_xy(copyright_x, term->get_height() - 3, "Copyright (C) 2011  Fabien LOISON");
		usleep(100 * 1000);
	}
	usleep(1000 * 1000);
	for (color=255 ; color>=232 ; color--)
	{
		term->set_fg_color(color);
		term->print_xy(banner_x, banner_y + 0, "  ___  ____  ____   ___  __    ____  ____  ");
		term->print_xy(banner_x, banner_y + 1, " / __)(_  _)(  _ \\ / __)(  )  ( ___)(  _ \\ ");
		term->print_xy(banner_x, banner_y + 2, "( (__  _)(_  )   /( (__  )(__  )__)  )(_) )");
		term->print_xy(banner_x, banner_y + 3, " \\___)(____)(_)\\_) \\___)(____)(____)(____/ ");
		term->print_xy(copyright_x, term->get_height() - 3, "Copyright (C) 2011  Fabien LOISON");
		usleep(100 * 1000);
	}
	term->draw_rectangle(0, 0, term->get_width(), term->get_height());
	term->set_fg_color(COLOR_WHITE);
}


/**
 * Draw the UI.
 *
 * Args:
 *     term -- The Terminal instance.
 */
void draw_ui(Terminal *term)
{
	term->set_bg_color(COLOR_UI_BG);
	term->draw_rectangle(0, 0, term->get_width(), term->get_height());
	term->set_bg_color(COLOR_UI);
	term->draw_rectangle(0, 0, term->get_width(), 1);
	term->draw_rectangle(0, term->get_height()-1,
			term->get_width(), term->get_height());
	term->set_fg_color(242);
	term->print_xy(1, 0, "Circled ");
	cout << VERSION << flush;
}


/**
 * Display an help message at the bottom of the screen.
 *
 * Args:
 *     hint -- The message to display.
 *     term -- The Terminal instance.
 */
void display_hint(char *hint, Terminal *term)
{
	term->set_bg_color(COLOR_UI);
	term->set_fg_color(COLOR_WHITE);
	term->draw_rectangle(0, term->get_height() - 1,
			term->get_width(),  term->get_height());
	term->move_xy(1, term->get_height() - 1);
	cout << hint << flush;
}


/**
 * Ask for the board size.
 *
 * Args:
 *     term -- The terminal instance;
 *
 * Returns:
 *     BOARD_SIZE_SMALL, BOARD_SIZE_MEDIUM or BOARD_SIZE_LARGE.
 */
int get_board_size(Terminal *term)
{
	draw_ui(term);
	display_hint("Select the size of the board.", term);
	term->set_title("Circled [Menu]");

	int x = (term->get_width() - 40) / 2;
	int y = term->get_height() / 2;

	term->set_fg_color(COLOR_WHITE);

	//Button Small
	term->set_bg_color(COLOR_BLACK);
	term->draw_rectangle(x + 2, y - 5, x + 42, y - 5 + 3);
	term->set_bg_color(COLOR_UI);
	term->draw_text_rectangle(x, y - 6, x + 40, y - 6 + 3, "Small");
	MouseEventBox btn_small(x, y - 6, x + 40, y - 6 + 3);

	//Button Medium
	term->set_bg_color(COLOR_BLACK);
	term->draw_rectangle(x + 2, y, x + 42, y + 3);
	term->set_bg_color(COLOR_UI);
	term->draw_text_rectangle(x, y - 1, x + 40, y - 1 + 3, "Medium");
	MouseEventBox btn_medium(x, y - 1, x + 40, y - 1 + 3);

	//Button Large
	term->set_bg_color(COLOR_BLACK);
	term->draw_rectangle(x + 2, y + 5, x + 42, y + 5 + 3);
	term->set_bg_color(COLOR_UI);
	term->draw_text_rectangle(x, y + 4, x + 40, y + 4 + 3, "Large");
	MouseEventBox btn_large(x, y + 4, x + 40, y + 4 + 3);

	int choice = -1;

	Event ev;
	do {
		ev = term->get_event();
		if (btn_small.check_collide(&ev))
		{
			choice = BOARD_SIZE_SMALL;
		}
		else if (btn_medium.check_collide(&ev))
		{
			choice = BOARD_SIZE_MEDIUM;
		}
		else if (btn_large.check_collide(&ev))
		{
			choice = BOARD_SIZE_LARGE;
		}
	} while (choice == -1);

	return choice;
}


/**
 * Play... \o/
 *
 * Args:
 *     term -- The Terminal instance.
 */
void play(Terminal *term)
{
	//Board size
	int width = 6;
	int height = 5;
	switch (get_board_size(term))
	{
		case BOARD_SIZE_SMALL:
			width = 8;
			height = 7;
			break;
		case BOARD_SIZE_MEDIUM:
			width = 14;
			height = 13;
			break;
		case BOARD_SIZE_LARGE:
			width = 22;
			height = 17;
			break;
	}

	//Draw the UI
	draw_ui(term);
	term->print_xy(term->get_width() - 10, 0, "[q] Quit");

	//Initialize the game
	Game game(width, height, "Player 1", "Player 2", term);
	
	//Play :D
	while (game.in_game)
	{
		//Player 1
		term->set_title("Circled [Player 1]");
		if (game.players[0].move_piece(&game.board)) break;
		if (game.players[0].put_brick(&game.board)) break;

		game.check_end_of_game();
		if (!game.in_game)
		{
			break;
		}

		//Player 2
		term->set_title("Circled [Player 2]");
		if (game.players[1].move_piece(&game.board)) break;
		if (game.players[1].put_brick(&game.board)) break;

		game.check_end_of_game();
	}

	//Game aborted
	if (game.winner == -1)
	{
		return;
	}

	//There is a winner
	term->set_title("Circled [Game Over]");
	game.players[game.winner].display_hint(
			"Good game!   -- Press any key to continue --", term);

	term->set_bg_color(COLOR_UI);
	term->set_fg_color(COLOR_WHITE);

	term->draw_text_rectangle(0, term->get_height() / 2 - 2,
			term->get_width(), term->get_height() / 2 + 1,
			game.players[game.winner].get_nickname() + string(" win!"));

	//Pause
	term->get_event();
}
